#include "parallel.h"
#include "interface.h"

void parallel_cube_host_(complex *BRAU, complex *BRAL,
                         complex *STAIRS, short *FENO, int *FELI,
                         int *NELI, short *IBUSO, short *PNODIA, int *NBUS)
{
   static int              Virgin = TRUE;
   FILE                   *SystemInfoFile;
   char                    machine[INFO_STRLEN];

   static int               PQ;
   static int               P;
   static int               Q;
   static index             p;
   static index             q;

   static int               old_NBUS = 0;
   int                      M;
   int                      N;
   int                      m;
   int                      n;

   char                     Command[COMMAND_STRLEN];

   char                     Pivoting_strategy[INFO_STRLEN];
   static long              Strategy;

   static short             Row_Dis_Type;
   static char              Row_Extra_param1_name[2];
   static long              Row_Extra_param1;
   static char              Row_Extra_param2_name[2];
   static long              Row_Extra_param2;
   Cdistrib                *Row_Dis;
   void                    *Row_Extra = NULL;
   void                   (*Row_Distribution)();
   int                    (*Row_Inverse_Distribution)();
   int                    (*Row_Distribution_limit)();
   int                    (*Row_Distribution_init)();

   static short             Col_Dis_Type;
   static char              Col_Extra_param1_name[2];
   static long              Col_Extra_param1;
   static char              Col_Extra_param2_name[2];
   static long              Col_Extra_param2;
   Cdistrib                *Col_Dis;
   void                    *Col_Extra = NULL;
   void                   (*Col_Distribution)();
   int                    (*Col_Inverse_Distribution)();
   int                    (*Col_Distribution_limit)();
   int                    (*Col_Distribution_init)();

   static int               Nnodes;

   char                     node_pgm_name[PGM_STRLEN];

   static ZIP_ADDRESSEES   *Addr_List;
   static ZIP_MAILER       *Mailer;

   static matrix_host_bin   Matrix;
   struct element         **Pointer;
   int                     *int_Pointer;
   int                      i;
   int                      j;

   int                      Index;
   int                      I;
   int                      J;

   int                      JB;

   long                     sec;
   long                     usec;

   printf("cube host -- demo for parallel sparse LU solver\n");

   if (Virgin == TRUE)
   {
      Virgin = FALSE;

      strcpy(node_pgm_name, NODE_PGM_NAME);

      if ((SystemInfoFile = fopen("cube.sys", "r")) == NULL)
      {
         perror("parallel_cube_host()");
         exit(1);
      }

      fscanf(SystemInfoFile, "%s%*c", machine);
#ifdef DEBUG_parallel_cube_host
      printf("machine = %s\n", machine);
#endif

      fscanf(SystemInfoFile, "%d %d%*c", &P, &Q);
#ifdef DEBUG_parallel_cube_host
      printf("Logical 2D Grid dimensions: %d %d\n", P, Q);
#endif

      PQ = P * Q;

      sprintf(Command, "getcube %s %dn", machine, PQ);

      if (system(Command))
         ckill(-1, -1, "");

      fscanf(SystemInfoFile, "%s", Pivoting_strategy);

      if (strstr(Pivoting_strategy, "_PARTIAL_") != NULL)
      {
         if (strstr(Pivoting_strategy, "_ROW_") != NULL)
         {
            Strategy = CMATRIX_PARTIAL_ROW_PIVOTING;
            printf("CMATRIX_PARTIAL_ROW_PIVOTING\n");
         }
         else if (strstr(Pivoting_strategy, "_COLUMN_") != NULL)
         {
            Strategy = CMATRIX_PARTIAL_COLUMN_PIVOTING;
            printf("CMATRIX_PARTIAL_COLUMN_PIVOTING\n");
         }
      }
      else if (strstr(Pivoting_strategy, "_MULTI_") != NULL)
      {
         if (strstr(Pivoting_strategy, "_ROW_") != NULL)
         {
            Strategy = CMATRIX_MULTI_ROW_PIVOTING;
            printf("CMATRIX_MULTI_ROW_PIVOTING\n");
         }
         else if (strstr(Pivoting_strategy, "_COLUMN_") != NULL)
         {
            Strategy = CMATRIX_MULTI_COLUMN_PIVOTING;
            printf("CMATRIX_MULTI_COLUMN_PIVOTING\n");
         }
      }
      else if (strstr(Pivoting_strategy, "_NO_") != NULL)
      {
         Strategy = CMATRIX_NO_PIVOTING;
         printf("CMATRIX_NO_PIVOTING\n");
      }
      else if (strstr(Pivoting_strategy, "_COMPLETE_") != NULL)
      {
         Strategy = CMATRIX_COMPLETE_PIVOTING;
         printf("CMATRIX_COMPLETE_PIVOTING\n");
      }
      else if (strstr(Pivoting_strategy, "_PRESET_") != NULL)
      {
         Strategy = CMATRIX_PRESET_PIVOTING;
         printf("CMATRIX_PRESET_PIVOTING\n");
      }
      else if (strstr(Pivoting_strategy, "_MARKOWITZ_") != NULL)
      {
         Strategy = CMATRIX_MARKOWITZ_COUNT_PIVOTING;
         printf("CMATRIX_MARKOWITZ_COUNT_PIVOTING\n");
      }

      read_distribution(SystemInfoFile, &Row_Dis_Type,
                        Row_Extra_param1_name, &Row_Extra_param1,
                        Row_Extra_param2_name, &Row_Extra_param2);

      read_distribution(SystemInfoFile, &Col_Dis_Type,
                        Col_Extra_param1_name, &Col_Extra_param1,
                        Col_Extra_param2_name, &Col_Extra_param2);

      fclose(SystemInfoFile);

      if (Zip_init() == NULL)
      {
         printf("parallel_cube_host: error - zip_init\n");
         exit(1);
      }

      Nnodes = zip_nnodes();

      print("csparsetest1: Nnodes = %d vs PQ = %d\n", Nnodes, PQ);

      Addr_List = zip_new_cohort(PQ, COHORT_FIRST_NODE, COHORT_PID, TRUE);

#ifdef DEBUG_parallel_cube_host
      print("etest1: checkpoint 0\n");
#endif

      /* spawn cohort of processes: */
      Zip_spawn(node_pgm_name, Addr_List);

#ifdef DEBUG_parallel_cube_host
      print("parallel_cube_host: checkpoint 0.1\n");
#endif

      /* mailer for cohort: */
      if ((Mailer = g2_grid_open(&P, &Q, Addr_List)) == NULL)
      {
          printf("csparsetest1: error - failed to open grid mailer\n");
          exit(1);
      }

#ifdef DEBUG_parallel_cube_host
      print("csparsetest1: checkpoint 0.2\n");
#endif

      Zip_push(Mailer);

      G2_pq(p, q);

#ifdef DEBUG_parallel_cube_host
      print("csparsetest1: checkpoint 1\n");
#endif

      transmit_intro_letter(Mailer, Row_Dis_Type, Col_Dis_Type,
                            Row_Extra_param1, Row_Extra_param2,
                            Col_Extra_param1, Col_Extra_param2, Strategy);

      Matrix.P = P;
      Matrix.Q = Q;

      if ((Matrix.A = (element ***) malloc(P * sizeof(element **))) == NULL)
      {
         printf("parallel_cube_host(): Out of memory!\n");
         exit(1);
      }

      if ((Pointer = (element **) malloc(PQ * sizeof(element *))) == NULL)
      {
         printf("parallel_cube_host(): Out of memory!\n");
         exit(1);
      }

      for (i = 0, Index = 0; i < P; i++, Index += Q)
      {
         (Matrix.A)[i] = &(Pointer[Index]);

         for (j = 0; j < Q; j++)
            (Matrix.A)[i][j] = NULL;
      }

      if ((Matrix.A_element_Count = (int **) malloc(P * sizeof(int *))) == NULL)
      {
         printf("parallel_cube_host(): Out of memory!\n");
         exit(1);
      }

      if ((int_Pointer = (int *) malloc(PQ * sizeof(int))) == NULL)
      {
         print("parallel_cube_host(): Out of memory!\n");
         exit(1);
      }

      for (i = 0, Index = 0; i < P; i++, Index += Q)
      {
         (Matrix.A_element_Count)[i] = &(int_Pointer[Index]);

         for (j = 0; j < Q; j++)
            (Matrix.A_element_Count)[i][j] = 0;
      }

      Matrix.Pool = NULL;
   }

   if (old_NBUS != *NBUS)
   {
      if (old_NBUS != 0)
      {
         Free(Row_Dis -> data);
         Free(Row_Dis);
         Free(Row_Extra);
         Free(Col_Dis -> data);
         Free(Col_Dis);
         Free(Col_Extra);
      }

      Matrix.M = M = *NBUS * 2;
      set_distribution(Mailer, Row_Dis_Type, CDISTRI_ROW, &Row_Dis,
                       &Row_Distribution, &Row_Inverse_Distribution,
                       &Row_Distribution_limit, &Row_Distribution_init,
                       P, p, M, &m, (int) Row_Extra_param1,
                       (int) Row_Extra_param2, &Row_Extra);

      Matrix.Row_Distribution = Row_Distribution;
      Matrix.Row_Extra = Row_Extra;

      Matrix.N = N = *NBUS * 2;
      set_distribution(Mailer, Col_Dis_Type, CDISTRI_COLUMN, &Col_Dis,
                       &Col_Distribution, &Col_Inverse_Distribution,
                       &Col_Distribution_limit, &Col_Distribution_init,
                       Q, q, N, &n, (int) Col_Extra_param1,
                       (int) Col_Extra_param2, &Col_Extra);

      Matrix.Col_Distribution = Col_Distribution;
      Matrix.Col_Extra = Col_Extra;
   }

#ifdef DEBUG_parallel_cube_host
   print("parallel_cube_host: Matrix data\n");
#endif

   for (i = 0; i < *NBUS - 1; i++)
   {
      I = (int) IBUSO[i] - 1;

#ifdef DEBUG_parallel_cube_host
      printf("I = %d\n", I);

      print("A[%d][%d] = (%g, %g)\n", I, I, STAIRS[I].real,
             STAIRS[I].imag);
#endif
      Save_matrix_data_diagonal(I, STAIRS[I], &Matrix);

      Index = FELI[I] - 1;
      J = FENO[Index] - 1;

#ifdef DEBUG_parallel_cube_host
      printf("J = %d\n", J);
#endif

      while (J >= 0)
      {
         if (PNODIA[J] > PNODIA[I])
         {
            JB = (Index >> 1);

            if (I < J)
            {
#ifdef DEBUG_parallel_cube_host
               print("A[%d][%d] = (%g, %g)\n", I, J, BRAU[JB].real,
                     BRAU[JB].imag);
               print("A[%d][%d] = (%g, %g)\n", J, I, BRAL[JB].real,
                     BRAL[JB].imag);
#endif
               Save_matrix_data_off_diagonal(I, J, BRAU[JB], BRAL[JB], &Matrix);
            }
            else
            {
#ifdef DEBUG_parallel_cube_host
               print("A[%d][%d] = (%g, %g)\n", I, J, BRAL[JB].real,
                     BRAL[JB].imag);
               print("A[%d][%d] = (%g, %g)\n", J, I, BRAU[JB].real,
                     BRAU[JB].imag);
#endif
               Save_matrix_data_off_diagonal(I, J, BRAL[JB], BRAU[JB], &Matrix);
            }
         }
         Index = NELI[Index] - 1;
         J = FENO[Index] - 1;
      }
   }

   I = (int) IBUSO[*NBUS - 1] - 1;

#ifdef DEBUG_parallel_cube_host
   print("A[[%d][%d] = (%g, %g)\n", I, I, STAIRS[I].real, STAIRS[I].imag);
#endif

   Save_matrix_data_diagonal(I, STAIRS[I], &Matrix);

   transmit_matrix_letter(Mailer, &Matrix);

   clear_matrix(&Matrix);

   receive_timing_letter(Mailer, &sec, &usec);

   print("Timing Data: %ld sec %ld usec\n", sec, usec);

#if 0
   Zip_pop();
   zip_exit();

   system("freecube");
#endif
}

void read_distribution(FILE *InputFile, short *Dis_Type,
                       char *Extra_param1_name, long *Extra_param1,
                       char *Extra_param2_name, long *Extra_param2)
{
   char Dis_Type_name[INFO_STRLEN];

   fscanf(InputFile, "%s", Dis_Type_name);
   printf("Dis_Type_name = %s\n", Dis_Type_name);

   /* defaults */
   strcpy(Extra_param1_name, "*\0");
   *Extra_param1 = 0;

   strcpy(Extra_param2_name, "*\0");
   *Extra_param2 = 0;

   if (strstr(Dis_Type_name, "CDISTRI_LINEAR"))
   {
      *Dis_Type = CDISTRI_LINEAR;
   }
   else if (strstr(Dis_Type_name, "CDISTRI_SCATTERED"))
   {
      *Dis_Type = CDISTRI_SCATTERED;
   }
   else if (strstr(Dis_Type_name, "CDISTRI_BLINEAR"))
   {
      *Dis_Type = CDISTRI_BLINEAR;

      fscanf(InputFile, "%ld", Extra_param1);
      printf("Row data distribution Block (B) factor = %ld",
             *Extra_param1);

      strcpy(Extra_param1_name, "B\0");
   }
   else if (strstr(Dis_Type_name, "CDISTRI_BSCATTERED"))
   {
      *Dis_Type = CDISTRI_BSCATTERED;

      fscanf(InputFile, "%ld", Extra_param1);
      printf("Row data distribution Block (B) factor = %ld", *Extra_param1);

      strcpy(Extra_param1_name, "B\0");
   }
   else if (strstr(Dis_Type_name, "CDISTRI_GBLINEAR"))
   {
      *Dis_Type = CDISTRI_GBLINEAR;

      fscanf(InputFile, "%ld", Extra_param1);
      printf("Row data distribution Block (B) factor = %ld",
             *Extra_param1);

      strcpy(Extra_param1_name, "B\0");
   }
   else if (strstr(Dis_Type_name, "CDISTRI_GBSCATTERED"))
   {
      *Dis_Type = CDISTRI_GBSCATTERED;

      fscanf(InputFile, "%ld", Extra_param1);
      printf("Row data distribution Block (B) factor = %ld",
             *Extra_param1);

      strcpy(Extra_param1_name, "B\0");
   }
   else if (strstr(Dis_Type_name, "CDISTRI_G1PARAM"))
   {
      *Dis_Type = CDISTRI_G1PARAM;

      fscanf(InputFile, "%ld", Extra_param2);
      printf("Row data distribution Scattering (S) Parameter = %ld",
             *Extra_param2);

      strcpy(Extra_param2_name, "S\0");
   }
   else if (strstr(Dis_Type_name, "CDISTRI_G2PARAM"))
   {
      *Dis_Type = CDISTRI_G2PARAM;

      fscanf(InputFile, "%ld", Extra_param1);
      printf("Data distribution Block (B) factor = %ld", *Extra_param1);

      strcpy(Extra_param1_name, "B\0");

      fscanf(InputFile, "%ld", Extra_param2);
      printf("Row data distribution Scattering (S) Parameter = %ld",
             *Extra_param2);

      strcpy(Extra_param2_name, "S\0");
   }
}

void Save_matrix_data_diagonal(int I, complex Diag_Value, matrix_host_bin *Matrix)
{
   char      *Subroutine_name = "Save_matrix_data_diagonal():";

   int        M = Matrix -> M;
   int        N = Matrix -> N;

   int        P = Matrix -> P;
   int        Q = Matrix -> Q;

   void     (*Row_Distribution)();
   void     (*Col_Distribution)();
   void      *Row_Extra = Matrix -> Row_Extra;
   void      *Col_Extra = Matrix -> Col_Extra;

   element ***A = Matrix -> A;
   int      **A_element_Count = Matrix -> A_element_Count;

   element   *Pool = Matrix -> Pool;

   int        Offset;
   int        Inc;
   index      p;
   index      q;
   index      i;
   index      j;

   element   *new_element;

   Row_Distribution = Matrix -> Row_Distribution;
   Col_Distribution = Matrix -> Col_Distribution;

#if 1
   /* four quater partition */
   /* A -B */
   /* B A */
   Offset = M >> 1;
#else
   
#endif

   (*Row_Distribution)(I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Diag_Value.real);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef DEBUG_Save_matrix_data_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, I, I, Diag_Value.real);
#endif

   (*Row_Distribution)(I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(Offset + I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, -Diag_Value.imag);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef DEBUG_Save_matrix_data_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, I, Offset + I, -Diag_Value.imag);
#endif

   (*Row_Distribution)(Offset + I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Diag_Value.imag);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef DEBUG_Save_matrix_data_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, Offset + I, I, Diag_Value.imag);
#endif

   (*Row_Distribution)(Offset + I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(Offset + I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Diag_Value.real);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef DEBUG_Save_matrix_data_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, Offset + I, Offset + I, Diag_Value.real);
#endif

   Matrix -> Pool = Pool;
}

void Save_matrix_data_off_diagonal(int I, int J, complex Upper_Value,
                                   complex Lower_Value, matrix_host_bin *Matrix)
{
   char      *Subroutine_name = "Save_matrix_data_off_diagonal():";

   int        M = Matrix -> M;
   int        N = Matrix -> N;

   int        P = Matrix -> P;
   int        Q = Matrix -> Q;

   void     (*Row_Distribution)();
   void     (*Col_Distribution)();
   void      *Row_Extra = Matrix -> Row_Extra;
   void      *Col_Extra = Matrix -> Col_Extra;

   element ***A = Matrix -> A;
   int      **A_element_Count = Matrix -> A_element_Count;

   element   *Pool = Matrix -> Pool;

   int        Offset;
   index      p;
   index      q;
   index      i;
   index      j;

   element    *new_element;

   Row_Distribution = Matrix -> Row_Distribution;
   Col_Distribution = Matrix -> Col_Distribution;

   Offset = M >> 1;

   (*Row_Distribution)(I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(J, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Upper_Value.real);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, I, J, Upper_Value.real);
#endif

   (*Row_Distribution)(J, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Lower_Value.real);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, J, I, Lower_Value.real);
#endif

   (*Row_Distribution)(I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(Offset + J, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, -Upper_Value.imag);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, I, Offset + J, -Upper_Value.imag);
#endif

   (*Row_Distribution)(J, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(Offset + I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, -Lower_Value.imag);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, J, Offset + I, -Lower_Value.imag);
#endif

   (*Row_Distribution)(Offset + I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(J, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Upper_Value.imag);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n",  p, q,
         i, j, Offset + I, J, Upper_Value.imag);
#endif

   (*Row_Distribution)(Offset + J, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Lower_Value.imag);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, Offset + J, I, Lower_Value.imag);
#endif

   (*Row_Distribution)(Offset + I, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(Offset + J, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Upper_Value.real);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, Offset + I, Offset + J, Upper_Value.real);
#endif

   (*Row_Distribution)(Offset + J, &p, &i, P, M, Row_Extra);
   (*Col_Distribution)(Offset + I, &q, &j, Q, N, Col_Extra);
   new_element = (element *) get_new_element(&Pool, i, j, Lower_Value.real);
   new_element -> Next = A[p][q];
   A[p][q] = new_element;
   A_element_Count[p][q]++;

#ifdef Save_matrix_data_off_diagonal
   print("(p, q, i, j) = (%hd, %hd, %hd, %hd) Real A[%d][%d] = %lf\n", p, q, i,
         j, Offset + J, Offset + I, Lower_Value.real);
#endif

   Matrix -> Pool = Pool;
}

element *get_new_element(element **Pool, index i, index j, double Value)
{
   char    *Subroutine_name = "get_new_element():";
   element *new_element;

   if (*Pool != NULL)
   {
      new_element = *Pool;
      *Pool = new_element -> Next;
   }
   else
   {
      new_element = (element *) malloc(sizeof(element));
   }

   if (new_element == NULL)
   {
      print("%s Out of memory!\n", Subroutine_name);
      exit(1);
   }
   else
   {
      new_element -> i = i;
      new_element -> j = j;
      new_element -> Value = Value;
      new_element -> Next = NULL;
   }

   return(new_element);
}

void clear_matrix(matrix_host_bin *Matrix)
{
   char      *Subroutine_name = "clear_matrix():";
   int        p;
   int        q;
   int        P = Matrix -> P;
   int        Q = Matrix -> Q;
   element   *Pool = Matrix -> Pool;
   element   *Pointer;
   element   *Trialer;
   element ***A = Matrix -> A;
   int      **A_element_Count = Matrix -> A_element_Count;
   int        Check_Count;

   for (p = 0; p < P; p++)
      for (q = 0; q < Q; q++)
         if ((Pointer = A[p][q]) != NULL)
         {
            Check_Count = 0;
            do
            {
               Check_Count++;
               Trialer = Pointer;
               Pointer = Pointer -> Next;
            } while (Pointer != NULL);

            Trialer -> Next = Pool; 
            Pool = A[p][q];
            A[p][q] = NULL;

            if (Check_Count != A_element_Count[p][q])
            {
               print("%s Memory lost at (%d, %d)!\n", Subroutine_name, p, q);
               print("Check_Count = %d, A_element_Count[%d][%d] = %d\n",
                     Check_Count, p, q, A_element_Count[p][q]);
            }

            A_element_Count[p][q] = 0;
         }
}
